/*Property of Four Js*
* (c) Copyright Four Js 1995, 2016. All Rights Reserved.
* * Trademark of Four Js Development Tools Europe Ltd
*   in the United States and elsewhere
* 
* Four Js and its suppliers do not warrant or guarantee that these
* samples are accurate and suitable for your purposes. Their inclusion is
* purely for information purposes only.
*/
/* Barcode Scanner Extension with the IOS native classes*/
#import <Foundation/Foundation.h>
#import <UIKit/UIKit.h>
/* GMI specific headerfile */
#import <gmi/frontcall.h>
@import AVFoundation;
@class GMIScanBarCodeViewController;
@interface GMIScanBarCode : FrontCall<AVCaptureMetadataOutputObjectsDelegate>
{
  GMIScanBarCodeViewController *_vc;
  bool _detectionFinished;
}
-(void)endWithCode:(NSString*)code andType:(NSString*)type error:(NSError*)error;
@end

@interface GMIScanBarCodeViewController:UIViewController
{
  AVCaptureSession* _session;
  AVCaptureVideoPreviewLayer *_prevLayer;
  UIView *_highlightView;
  UILabel *_label;
  UIButton *_button;
  GMIScanBarCode* __weak _delegate;
  NSError* _error;
}
@property (readonly) UILabel* label;
@property (readonly) UIView* highlightView;
@property (readonly) AVCaptureVideoPreviewLayer * prevLayer;
@property (readonly) NSError* error;
@end

@interface UIView(GMIScanLayoutHelpers)
-(void)addLayoutSubview:(UIView *)subview;
-(void)topAlignsTop:(UIView*)view1 spacing:(CGFloat)spacing;
-(void)bottomAlignsBottom:(UIView*)view1 spacing:(CGFloat)spacing;
-(void)leftAlignsLeft:(UIView*)view1 spacing:(CGFloat)spacing;
-(void)rightAlignsRight:(UIView*)view1 spacing:(CGFloat)spacing;
@end

@implementation GMIScanBarCodeViewController
-(id) initWithDelegate:(GMIScanBarCode*)delegate
{
  self = [super init];
  if (self !=nil ) {
    _delegate=delegate;
  }
  return self;
}

-(void)setOrientationAndBounds
{
  UIDeviceOrientation orientation = [UIDevice currentDevice].orientation;
  [_prevLayer.connection setVideoOrientation:(AVCaptureVideoOrientation)orientation];
  _prevLayer.frame = self.view.bounds;
}

-(void)viewDidLoad
{
  [super viewDidLoad];
  UIView* view=self.view;
  _highlightView = [[UIView alloc] init];
  _highlightView.layer.borderColor = [UIColor greenColor].CGColor;
  _highlightView.layer.borderWidth = 3;
  [view addSubview : _highlightView];
  _label = [[UILabel alloc] init];
  _label.backgroundColor = [UIColor colorWithWhite : 0.15 alpha : 0.65];
  _label.textColor = [UIColor whiteColor];
  _label.textAlignment = NSTextAlignmentCenter;
  _label.text = @ "(No code detected yet)";
  _label.numberOfLines=1; //otherwise multiple lines may cause overflows
  [_label setPreferredMaxLayoutWidth:view.bounds.size.width];
  [view addLayoutSubview : _label];
  [view leftAlignsLeft:_label spacing:0];
  [view rightAlignsRight:_label spacing:0];
  [view bottomAlignsBottom:_label spacing:0];
  _button=[[UIButton alloc] init];
  [_button setTitle:@"Cancel" forState:UIControlStateNormal];
  [_button addTarget:_delegate action:@selector(cancel) forControlEvents:UIControlEventTouchUpInside];
  [_button setTitleColor:[UIColor blueColor] forState:UIControlStateNormal];
  [view addLayoutSubview : _button];
  [view leftAlignsLeft:_button spacing:20];
  [view topAlignsTop:_button spacing:20];
  _session = [[AVCaptureSession alloc] init];
  AVCaptureDevice* device = [AVCaptureDevice defaultDeviceWithMediaType : AVMediaTypeVideo];
  NSError* error=nil;
  AVCaptureDeviceInput* input = [AVCaptureDeviceInput deviceInputWithDevice : device error : &error];
  if( input == nil ) {
    _error=error;
    return;
  }
  [_session addInput : input];
  AVCaptureMetadataOutput* output = [[AVCaptureMetadataOutput alloc] init];
  [output setMetadataObjectsDelegate : _delegate queue : dispatch_get_main_queue()];
  [_session addOutput : output];

  output.metadataObjectTypes = [output availableMetadataObjectTypes];
  _prevLayer = [AVCaptureVideoPreviewLayer layerWithSession : _session];
  [self setOrientationAndBounds];
  _prevLayer.videoGravity = AVLayerVideoGravityResizeAspectFill;
  [view.layer addSublayer : _prevLayer];
  [_session startRunning];
  [view bringSubviewToFront:_label];
  [view bringSubviewToFront:_highlightView];
  [view bringSubviewToFront:_button];
}

//handle rotation
- (void)viewWillTransitionToSize:(CGSize)size
       withTransitionCoordinator:(id<UIViewControllerTransitionCoordinator>)coordinator
{
  [super viewWillTransitionToSize:size withTransitionCoordinator:coordinator];
  [coordinator animateAlongsideTransition:nil completion:^(id<UIViewControllerTransitionCoordinatorContext> _Nonnull context) {
    [self setOrientationAndBounds];
  }];
}

-(void)dealloc
{
  [_session stopRunning];
}

@end

/* return an array of your FrontCall classes, each module needs a distinct class */
NSArray* frontCalls()
{
  return @[ [GMIScanBarCode class] ];
}

@implementation GMIScanBarCode
- (NSString*) moduleName
{
  return @"gmiscan";
}

-(void)scanBarCode
{
  _detectionFinished=false;
  _didSetResult = true;
  //first check the camer permissions
  AVAuthorizationStatus authStatus = [AVCaptureDevice authorizationStatusForMediaType:AVMediaTypeVideo];
  if(authStatus == AVAuthorizationStatusDenied){
    [self errorWithMessage:@"access denied"];
    return;
  } else if(authStatus == AVAuthorizationStatusNotDetermined){
    GMIScanBarCode* __weak weakSelf = self;
    [AVCaptureDevice requestAccessForMediaType:AVMediaTypeVideo
                             completionHandler:^(BOOL granted) {
                               dispatch_async (dispatch_get_main_queue (), ^{
                                 if(granted){
                                   [weakSelf showController];
                                 } else {
                                   [weakSelf errorWithMessage:@"access denied"];
                                 }
                               });
                             }];
  } else {
    [self showController];
  }
}

-(void)showController
{
  _vc=[[GMIScanBarCodeViewController alloc] initWithDelegate:self];
  [self.rootViewController presentViewController:_vc animated:YES completion:^(void) {
    if (_vc.error!=nil) {
      [self endWithCode:nil andType:@"error" error:_vc.error];
    }
  }];
}

-(void)endWithCode:(NSString*)code andType:(NSString*)type error:(NSError*)error
{
  [self.rootViewController
   dismissViewControllerAnimated:TRUE
   completion:^(void) {
     if (error!=nil) {
       [self errorWithMessage:error.localizedDescription];
     } else {
       [self startResult];
       [self addStringResult:code];
       [self addStringResult:type];
       [self endResult];
     }
     _vc=nil;
  }];
}

-(void)cancel
{
  _detectionFinished=true;
  [self endWithCode:nil andType:@"canceled" error:nil];
}

static NSDictionary *s_barCodeDict=nil;
-(void)initDict
{
  if (s_barCodeDict!=nil) {return;}
  //init with the standard genero names
  s_barCodeDict=
  @{ AVMetadataObjectTypePDF417Code:@"PDF417",
     AVMetadataObjectTypeUPCECode:@"UPC-E",
     AVMetadataObjectTypeCode39Code:@"Code-39",
     AVMetadataObjectTypeCode39Mod43Code:@"Code-39-Mod43",
     AVMetadataObjectTypeEAN13Code:@"EAN-13",
     AVMetadataObjectTypeEAN8Code:@"EAN-8",
     AVMetadataObjectTypeCode93Code:@"Code-93",
     AVMetadataObjectTypeCode128Code:@"Code-128",
     AVMetadataObjectTypeQRCode:@"QR-Code",
     AVMetadataObjectTypeAztecCode:@"AZTEC",
     AVMetadataObjectTypeITF14Code:@"ITF-14",
     AVMetadataObjectTypeDataMatrixCode:@"DATA-MATRIX",
     AVMetadataObjectTypeInterleaved2of5Code:@"I2/5"
     };
}

-(NSArray*)barCodeTypes
{
  return s_barCodeDict.allKeys;
}

- (void)captureOutput:(AVCaptureOutput *)captureOutput
didOutputMetadataObjects:(NSArray *)metadataObjects
       fromConnection:(AVCaptureConnection *)connection
{
  if (_detectionFinished) {
    return;
  }
  [self initDict];
  AVMetadataMachineReadableCodeObject *barCodeObject;
  for (AVMetadataObject *metadata in metadataObjects) {
    for (NSString *type in [self barCodeTypes]) {
      if ([metadata.type isEqualToString:type])
      {
        barCodeObject = (AVMetadataMachineReadableCodeObject *)[_vc.prevLayer transformedMetadataObjectForMetadataObject:(AVMetadataMachineReadableCodeObject *)metadata];
        _vc.highlightView.frame = barCodeObject.bounds;
        NSString* detectionString = [(AVMetadataMachineReadableCodeObject *)metadata stringValue];
        _vc.label.text = detectionString; //or just @"SUCCESS";
        _detectionFinished=true;
        //just give the end user the chance to see which block was selected
        //choose whatever time you find appropriate
        dispatch_after(dispatch_time(DISPATCH_TIME_NOW, 0.5 * NSEC_PER_SEC), dispatch_get_main_queue(), ^(void){
          [self endWithCode:detectionString andType:s_barCodeDict[type] error:nil];
        });
        return;
      }
    }
  }
}

-(void)execute:(NSString*)name retCount:(int)retCount params:(NSArray*)params
{
  name=[name lowercaseString];
  [super execute:name retCount:retCount params:params];
  if ([name isEqualToString:@"scanbarcode"]) {
    [self scanBarCode];
  }
}

-(UIViewController*)rootViewController
{
  UIViewController* r=[super rootViewController];
  if (r!=nil) {return r;}
  //this is a kinda hack because self.rootViewController does not work yet
  UIWindow* win=UIApplication.sharedApplication.keyWindow;
  UIViewController* vc=win.rootViewController;
  while (vc.presentingViewController!=nil) { //walk up the chain
    vc=vc.presentingViewController;
  }
  return vc;
}

@end

@implementation UIView(GMIScanLayoutHelpers)
-(void)addLayoutSubview:(UIView *)subview
{
  [subview setTranslatesAutoresizingMaskIntoConstraints:NO];
  [self addSubview:subview];
}

-(void)topAlignsTop:(UIView*)view1 spacing:(CGFloat)spacing
{
  [self addConstraint:[NSLayoutConstraint
                       constraintWithItem:view1
                       attribute:NSLayoutAttributeTop
                       relatedBy:NSLayoutRelationEqual
                       toItem:self
                       attribute:NSLayoutAttributeTop
                       multiplier:1
                       constant:spacing]];
}


-(void)bottomAlignsBottom:(UIView*)view1 spacing:(CGFloat)spacing
{
  [self addConstraint:[NSLayoutConstraint
                       constraintWithItem:view1
                       attribute:NSLayoutAttributeBottom
                       relatedBy:NSLayoutRelationEqual
                       toItem:self
                       attribute:NSLayoutAttributeBottom
                       multiplier:1
                       constant:-spacing]];
}

-(void)leftAlignsLeft:(UIView*)view1 spacing:(CGFloat)spacing
{
  [self addConstraint:[NSLayoutConstraint
                       constraintWithItem:view1
                       attribute:NSLayoutAttributeLeft
                       relatedBy:NSLayoutRelationEqual
                       toItem:self
                       attribute:NSLayoutAttributeLeft
                       multiplier:1
                       constant:spacing]];
}

-(void)rightAlignsRight:(UIView*)view1 spacing:(CGFloat)spacing
{
  [self addConstraint:[NSLayoutConstraint
                       constraintWithItem:view1
                       attribute:NSLayoutAttributeRight
                       relatedBy:NSLayoutRelationEqual
                       toItem:self
                       attribute:NSLayoutAttributeRight
                       multiplier:1
                       constant:-spacing]];
}
@end
